/********************************************************
 * Copyright 1993, COMPAQ Computer Corporation
 ********************************************************
 *
 * Title   : $RCSfile: xlong.h,v $
 *
 * Version : $Revision: 1.2 $
 *
 * Date    : $Date: 2001/11/09 03:23:08 $
 *
 * Author  : $Author: dwayne $
 *
 ********************************************************
 *
 * Change Log :
 *
 *           $Log: xlong.h,v $
 *           Revision 1.2  2001/11/09 03:23:08  dwayne
 *           re-formatted non-chim c files and headers
 *
 *           Revision 1.1  2001/10/30 20:46:47  dwayne
 *           Reorganized and restructured ad160 driver
 *
 *           Revision 1.1  2001/10/30 19:26:24  dwayne
 *           Imported ad160 driver for openserver
 *
 *           Revision 1.1.1.1  2001/10/26 22:23:41  dwayne
 *           Initial checkin of adaptec ultra 160
 *
 *           Revision 1.1.2.1  2000/04/21 19:40:32  adaptec
 *           CIM changes incorporated
 *
 *           Revision 1.3  1993/07/02 16:21:29  nickm
 *           Added shift right and shift left macros for xlong type.
 *           Added shift test code to test program main.
 *
 * Revision 1.2  1993/06/17  16:34:34  nickm
 * Simplified xlong definition.  Seems source compatible.
 *
 * Revision 1.1  1993/01/07  21:30:04  nickm
 * Initial revision
 *
 *           $EndLog$
 *
 ********************************************************/

/*
	The purpose for these macros is to allow a driver to
	add small positive integers to accumulators which may
	eventually become very very large.  Drivers are not
	permitted to use floating point.

	The type xlong_t is defined here and variables of this type
	will hold a 64-bit unsigned integer value.

	The following operations are defined mu macros for the
	type xlong_t:

	   In Kernel or User Code:
		Add an unsigned integer value to an xlong_t.
		Subtract an unsigned integer value from an xlong_t.
		Add an xlong_t to an xlong_t.
		Subtract an xlong_t from an xlong_t.
		Zero an xlong_t.
		Copy an xlong_t to an xlong_t.
	    Only in User Code:
		Convert an xlong_t to a double.

	The value of an xlong_t also may be assigned to another
	xlong_t using normal C syntax: a = b;  This is structure
	assignment.  If it makes you nervous, don't use it.
 */

/* A decicion was made to develop these macros specificly
   coded and tuned for 64-bit unsigned accumulators rather developing
   more generic macros which would handle accumulators of virtually
   any length.  A 64-bit unsigned number can represent over 10 billion
   billion (10 quintillion) and the operations are quite fast on 32-bit
   machines.  */

/* If stronger type shecking could be added at compile time,
   this could be helpful.  I didn't have time to figure a way
   to do that yet. */

#ifndef XLONG_DEF
#define XLONG_DEF

/* struct xlong_s { */
	union xlong_u {
		unsigned long ul[2];
		unsigned short us[4];
		unsigned char uc[8];
	};
/* }; */
/* typedef struct xlong_s xlong_t; */
typedef union xlong_u xlong_t;

/* These macro names were chosed to show the order of the arguments
   and the direction of operation (think COBOL) */

	/* any non-negative integer type is ok for int here */
#define add_int_to_xlong(vint,vxlong) { \
	register unsigned long vtemp = vxlong.ul[0]; \
	vxlong.ul[0] = vtemp + (vint); \
	if (vtemp > vxlong.ul[0]) \
		vxlong.ul[1]++; \
}

#define add_xlong_to_xlong(vxlong1,vxlong2) { \
	register unsigned long vtemp = vxlong2.ul[0]; \
	vxlong2.ul[0] = vtemp + vxlong1.ul[0]; \
	if (vtemp > vxlong2.ul[0]) \
		vxlong2.ul[1]++; \
	vxlong2.ul[1] += vxlong1.ul[1]; \
}

/*  Although SCO's documentation indicates that an unsigned long
    will be converted to a long before converting to double, the
    conversion of an unsigned long with the high order bit set
    is currently not failing to produce the desired result.
    If this changes, the high order bit of the low order unsigned
    long will need to be treated differently in xlong_to_double.
*/
	/* If float is used instead of double,
	   this will not work well.  */
#define xlong_to_double(vxlong,vdouble) \
    vdouble = vxlong.ul[0] + (vxlong.ul[1] * 4294967296.0);

#define zero_xlong(vxlong) \
	vxlong.ul[0] = 0; \
	vxlong.ul[1] = 0;

#define xlong_to_xlong(vxlong1,vxlong2) \
	vxlong2.ul[0] = vxlong1.ul[0]; \
	vxlong2.ul[1] = vxlong1.ul[1];

	/* any non-negative integer type is ok for int here */
#define subtract_int_from_xlong(vint,vxlong) { \
	register unsigned long vtemp = vxlong.ul[0]; \
	vxlong.ul[0] = vtemp - (vint); \
	if (vtemp < vxlong.ul[0]) \
		vxlong.ul[1]--; \
}

#define subtract_xlong_from_xlong(vxlong1,vxlong2) { \
	register unsigned long vtemp = vxlong2.ul[0]; \
	vxlong2.ul[0] = vtemp - vxlong1.ul[0]; \
	if (vtemp < vxlong2.ul[0]) \
		vxlong2.ul[1]--; \
	vxlong2.ul[1] -= vxlong1.ul[1]; \
}

#define xlong_shift_right(vxlong,vint) { \
	vxlong.ul[0] = (vxlong.ul[0] >> vint) | \
	(vxlong.ul[1] << (32 - vint)); \
	vxlong.ul[1] = vxlong.ul[1] >> vint; \
}

#define xlong_shift_left(vxlong,vint) { \
	vxlong.ul[1] = (vxlong.ul[1] << vint) | \
	(vxlong.ul[0] >> (32 - vint)); \
	vxlong.ul[0] = vxlong.ul[0] << vint; \
}


#ifdef TEST_XLONG
/*
	This is a small test which shows how the macros may be used.
	It also demonstrates that these operations are quite fast.
	They should not adversely affect driver performance.

	The test program xlong.c is:

		#define TEST_XLONG
		#include "xlong.h"

	The compile line is: make xlong.

	The run line is: nohup time ./xlong

	The result on my system is:

	    sizeof double 8
	    2 to the 8 is 256.000000
	    2 to the 16 is 65536.000000
	    2 to the 32 is 4294967296.000000
	    print_me is 63.000000
	    counter is 0 d1 is 6.300e+01 d2 is 0.000e+00
	    	difference is 6.300e+01 d3 is 6.300e+01
	    counter is 100000 d1 is 2.000e+14 d2 is 2.000e+14
	    	difference is 6.300e+01 d3 is 6.300e+01
	    counter is 200000 d1 is 4.000e+14 d2 is 4.000e+14
	    	difference is 6.300e+01 d3 is 6.300e+01
		    ...
	    counter is 4500000 d1 is 9.000e+15 d2 is 9.000e+15
	    	difference is 6.300e+01 d3 is 6.300e+01
	    counter is 4600000 d1 is 9.200e+15 d2 is 9.200e+15
	    	difference is 6.400e+01 d3 is 6.300e+01
		    ...
	    counter is 9900000 d1 is 1.980e+16 d2 is 1.980e+16
	   	 difference is 6.400e+01 d3 is 6.300e+01
	    counter is 10000000 d1 is 2.000e+16 d2 is 2.000e+16
	    	difference is 6.400e+01 d3 is 6.300e+01

	    real        8.0
	    user        7.8
	    sys         0.0
	
	The perfomance is thus about 20 million add_int_to_xlong
	operations in 8 seconds.  Very busy drivers might do this
	many operations in a day.
 */
main(int argc, char **argv)
{
	register j;
	int i;
	short inc_2;	/* 2 byte increment */
	unsigned long inc_4; /* 4 byte increment */
	double print_me_double;
	double d1, d2, d3;
	long counter = 0;
	xlong_t acc_1;	/* 8 byte accumulator */
	xlong_t acc_2;	/* 8 byte accumulator */
	xlong_t acc_3;	/* 8 byte accumulator */

	printf("sizeof double %d\n", sizeof(double));
	print_me_double = 256;
	printf("2 to the 8 is %f\n", print_me_double);
	print_me_double *= print_me_double;
	printf("2 to the 16 is %f\n", print_me_double);
	print_me_double *= print_me_double;
	printf("2 to the 32 is %f\n", print_me_double);
	print_me_double -= 1;
	inc_4 = print_me_double;
	printf("2 to the 32 -1 double to unsigned long is %u (%#x)\n",
		inc_4, inc_4);
	print_me_double = inc_4;
	printf("2 to the 32 -1 unsigned long to double is %f\n",
		print_me_double);
	print_me_double += 1;
	print_me_double *= print_me_double;
	printf("2 to the 64 is %f\n", print_me_double);
	inc_2 = 63;
	inc_4 = 2000000000;	/* two billion */
		/* zero two accumulators */
	zero_xlong(acc_1);
	zero_xlong(acc_2);
		/* add a two byte int to one accumulator */
	add_int_to_xlong(inc_2, acc_1);
		/* convert accumulator to double for printing */
	xlong_to_double(acc_1, print_me_double);
	printf("print_me is %f\n", print_me_double);

	printf("counter is %d ", counter);
	xlong_to_double(acc_1, d1);
	xlong_to_double(acc_2, d2);
		/* assignment from one accumulator to another */
	acc_3 = acc_1;
		/* subtraction modifies the second argument */
	subtract_xlong_from_xlong(acc_2, acc_3)
		/* subtract the converted values of acc_1 and 2 */
	print_me_double = d1 - d2;
		/* convert the differece between acc_1 and 2 */
	xlong_to_double(acc_3, d3);
	printf("d1 is %.3e d2 is %.3e difference is %.3e\ d3 is %.3e\n",
	    d1, d2, print_me_double, d3);
	    /* ten million times */
	for (i = 0; i < 100; i++)
	{
	    for (j = 0; j < 100000; j++)
	    {
		add_int_to_xlong(inc_4, acc_1);
		add_int_to_xlong(inc_4, acc_2);
	    }
	    counter += j;
	    printf("counter is %d ", counter);
	    xlong_to_double(acc_1, d1);
	    xlong_to_double(acc_2, d2);
		/* replace  acc_3 = acc_1; with a macro */
	    xlong_to_xlong(acc_1, acc_3);
	    subtract_xlong_from_xlong(acc_2, acc_3)
	    print_me_double = d1 - d2;
	    xlong_to_double(acc_3, d3);
		/* At about 9.0e+15 the difference in the doubles
		becomes less precise than the converted differece
		of the accumulators.  Since this is about nine thousand
		trillion, maybe you don't care unless you are in the
		kernel and can't do floating point. */
		/* Does this mean the U.S. national debt could be
		   counted in pennies in an xlong_t ? */
	    printf("d1 is %.3e d2 is %.3e difference is %.3e\ d3 is %.3e\n",
		d1, d2, print_me_double, d3);
	}
	acc_1.ul[0] = 0xffffffff;
	acc_1.ul[1] = 0xffffffff;
	xlong_to_double(acc_1, print_me_double);
	printf("2 to the 64 -1 xlong to double is %f or %.3e\n",
		print_me_double, print_me_double);
	i = 64;
	while (i >= 0)
	{
	    xlong_to_double(acc_1, print_me_double);
	    printf("2 to the %d -1 xlong to double is %f or %.3e\n",
		i, print_me_double, print_me_double);
	    i--;
	    xlong_shift_right(acc_1, 1);
	}
	i = 0;

	zero_xlong(acc_1);
	add_int_to_xlong(1, acc_1);
	while (i <= 64)
	{
	    xlong_to_double(acc_1, print_me_double);
	    printf("2 to the %d xlong to double is %f or %.3e\n",
		i, print_me_double, print_me_double);
	    i++;
	    xlong_shift_left(acc_1, 1);
	}
}
#endif /* #ifdef TEST_XLONG */
#endif /* #ifndef XLONG_DEF */
